// BWCryptoDocument.h -- a document that stores the cryptogram string and 
//                       the letter mapping dictionary

#import "BWCryptoDocument.h"
#import "BWCryptoView.h"
#import "NSDictionary+Letter.h"


// a category for forward references - when putting the method definition
// before its first use would be awkward or unstylish

@interface BWCryptoDocument(ForwardReferences)

- (void) updateLetterFrequencies;
- (void) resetTranslationToIdentity;
- (void) updateTranslatedLetter: (NSString *) letter
              forOriginalLetter: (NSString *) original;

@end // ForwardReferences




@implementation BWCryptoDocument

// make a new blank document

- (id) init
{
    if (self = [super init]) {

        // default by mapping A=A, B=B, and so on.
        letterMappingDict = [[NSMutableDictionary alloc] init];
        [letterMappingDict resetTranslationToIdentity];

        // The original "cryptogram", points folks on where to start
        originalString = @"WELCOME TO BORKWARE CRYPTOGRAMMER. TO BEGIN, CHOOSE \"ENTER CRYPTOGRAM\" AND TYPE IN YOUR PUZZLE.";
    }

    return (self);

} // init



// clean up our mess

- (void) dealloc
{
    [originalString release];
    [letterMappingDict release];

    [super dealloc];

} // dealloc



// the original string has changed.  Hang on to it, and tell the cryptoview
// that there's a new string to play with

- (void) changeOriginalString: (NSString *) string
{
    [originalString release];
    
    originalString = [string uppercaseString];
    [originalString retain];
    [cryptoView setOriginalString: originalString];

} // changeOriginalString



// this is the undo bottleneck for changing the cryptogram string

- (void) setOriginalString: (NSString *) string
{
    if (![string isEqualTo: originalString]) {

        NSUndoManager *undoManager = [self undoManager];
        
        [[undoManager prepareWithInvocationTarget: self]
            setOriginalString: [originalString copy]];
        
        [undoManager setActionName: @"Change Phrase"];
        
        [self changeOriginalString: string];
        
        [cryptoView setOriginalString: originalString];
        [self updateLetterFrequencies];
    }

} // setOriginalString



// NSDocument will use this nib file when creating new documents

- (NSString *) windowNibName
{
    return @"BWCryptoDocument";

} // windowNibName



// this is the 'awake from nib' for documents.

- (void) windowControllerDidLoadNib: (NSWindowController *) aController
{
    // make sure our superclass is all set up
    [super windowControllerDidLoadNib: aController];

    // set the cryptoview's string and mapping dict
    [cryptoView setOriginalString: originalString];
    [cryptoView setLetterMappingDict: letterMappingDict];

    // the frequencies matrix might not be 26 items (due to limitations
    // of interface builder - making it 26 long, you can't resize it
    // down small enough for a reasonably sized initial window.
    // So, we'll just keep adding rows until we hit 26, one for each letter

    int rowCount = [frequenciesMatrix numberOfRows];
    while (rowCount < 26) {
        [frequenciesMatrix addRow];
        rowCount++;
    }

    // populate the letter frequency matrix
    [self updateLetterFrequencies];

    // register to get 'letter changed' notifications from the view.
    // we probably could have used a delegate here instead of a notification,
    // but hey, this works

    NSNotificationCenter *defaultCenter;
    defaultCenter = [NSNotificationCenter defaultCenter];
    
    [defaultCenter addObserver: self
                   selector: @selector(handleLetterMapChangeNotification:)
                   name: BWCryptoView_LetterMapChangeNotification
                   object: cryptoView];

} // windowControllerDidLoadNib



// encode document data so it can be written to disk

- (NSData *) dataRepresentationOfType: (NSString *) aType
{
    // stick stuff into a dictionary.
    NSDictionary *dictionary;
    dictionary = [NSDictionary dictionaryWithObjectsAndKeys:
                                   originalString, @"string",
                               letterMappingDict, @"mappingDict", nil];

    // since everything is a property list class (strings and a dictionary
    // in our case), serialize it out to XML, to be sexy.

    NSData *data;
    data = [NSPropertyListSerialization
               dataFromPropertyList: dictionary
               format: NSPropertyListXMLFormat_v1_0
               errorDescription: nil];

    return (data);

} // dataRepresentationOfType



// document is being loaded from a file

- (BOOL) loadDataRepresentation: (NSData *) data  ofType: (NSString *) aType
{
    // turn the property list XML goodness into some property list classes
    NSDictionary *dictionary;
    dictionary = [NSPropertyListSerialization
                     propertyListFromData: data
                     mutabilityOption: NSPropertyListMutableContainers
                     format: nil
                     errorDescription: nil];

    // replace instance variables with the new stuff
    [originalString release];

    originalString = [dictionary objectForKey: @"string"];
    [originalString retain];

    [letterMappingDict release];

    letterMappingDict = [dictionary objectForKey: @"mappingDict"];
    [letterMappingDict retain];

    // this will only have an effect after the nib is loaded, so 
    // pretty much these will just have an effect when the user
    // chooses Revert

    [cryptoView setOriginalString: originalString];
    [cryptoView setLetterMappingDict: letterMappingDict];

    return (YES);

} // loadDataRepresentaiton


// there's not a method to just get our document's window, so dig
// around in the controllers and pick off the first one

- (NSWindow *) window
{
    NSArray *windowControllers;
    windowControllers = [self windowControllers];

    NSWindowController *controller;
    controller = [windowControllers objectAtIndex: 0];

    NSWindow *window;
    window = [controller window];

    return (window);

} // window


// structure for storing a letter and a count.  An array of these
// will be populated and sorted.

typedef struct LetterCount {
    unichar letter;
    int count;
} LetterCount;



// qsort comparison function.  Orders the items pointed to by count

static int compareLetterCounts (const void *thing1, const void *thing2)
{
    const LetterCount *t1, *t2;
    t1 = (const LetterCount *) thing1;
    t2 = (const LetterCount *) thing2;

    // small values, integer overflow is not a problem
    return (t2->count - t1->count);

} // compareLetterCounts


// walk the string, count up the number of each letter, then sort 'em
// and populate the letter frequency matrix

- (void) updateLetterFrequencies
{
    LetterCount count[26];

    // zero out the count, and set the letter for each thingie
    int i;
    for (i = 0; i < 26; i++) {
        count[i].letter = 'A' + i;
        count[i].count = 0;
    }

    NSString *string;
    string = [cryptoView originalString];

    // getting the length first lets us avoid a method call each loop
    int length;
    length = [string length];

    // walk the string
    for (i = 0; i < length; i++) {

        unichar letter;
        letter = [string characterAtIndex: i];

        // find the location in the array
        letter -= 'A';

        // a unichar is unsigned, so subtraction off the end will wrap
        // around to huge numbers.
        if (letter >= 26) {
            // skip non-letter junk
            continue;
        }

        count[letter].count++;
    }

    // sort the array by count
    qsort (count, 26, sizeof(LetterCount), compareLetterCounts);

    // walk the matrix and clear out all the cells.  In case there's
    // some leftover IB placeholders in them
    int row, column;
    for (row = 0; row < [frequenciesMatrix numberOfRows]; row++) {
        for (column = 0; column < [frequenciesMatrix numberOfColumns]; 
             column++) {

            NSCell *cell;
            cell = [frequenciesMatrix cellAtRow: row  column: column];
            [cell setStringValue: @""];
        }
    }

    int maxRow;
    maxRow = MIN(26, [frequenciesMatrix numberOfRows]);

    // walk the matrix and set the cells to match the letter:count
    // values we just calcualted

    for (i = 0; i < maxRow; i++) {
        NSCell *cell;

        NSString *letter;
        letter = [NSString stringWithCharacters: &count[i].letter  length: 1];

        NSString *map;
        map = [letterMappingDict translateLetter: letter];

        cell = [frequenciesMatrix cellAtRow: i  column: 0];
        [cell setStringValue: 
                  [NSString stringWithFormat: @"%@ (%@)",
                            letter, map]];

        string = [NSString stringWithFormat: @"%d", count[i].count];
        cell = [frequenciesMatrix cellAtRow: i  column: 1];
        [cell setStringValue: string];
    }

} // updateLetterFrequencies


// permute the letters.  Handy for re-starting the cryptogram

- (IBAction) scrambleLetters: (id) sender
{
    // seed the random number generator
    srandom (time(NULL));

    NSUndoManager *undoManager;
    undoManager = [self undoManager];

    [undoManager beginUndoGrouping];

    int i;
    for (i = 0; i < 100; i++) {
        unichar randomLetter;
        randomLetter = (random () % 26) + 'A';

        NSString *originalLetter;
        originalLetter = [NSString stringWithCharacters: &randomLetter
                                   length: 1];

        randomLetter = (random () % 26) + 'A';
        NSString *translatedLetter;
        translatedLetter = [NSString stringWithCharacters: &randomLetter
                                     length: 1];
        
        [self updateTranslatedLetter: translatedLetter
              forOriginalLetter: originalLetter];
    }
    [undoManager setActionName: @"Scramble Letters"];

    [undoManager endUndoGrouping];
    [self updateLetterFrequencies];

} // scrambleLetters



// move the transmuted letters from the replacement letters into
// the original string.  Also handy for creating new cryptograms.

- (IBAction) copyReplacementLettersToCryptogram: (id) sender
{
    int length;
    length = [originalString length];

    NSMutableString *string;
    string = [[NSMutableString alloc] initWithCapacity: length];
    [string autorelease];

    int i;
    for (i = 0; i < length; i++) {
        unichar character;
        character = [originalString characterAtIndex: i];
        
        NSString *letter;
        letter = [NSString stringWithCharacters: &character  length: 1];

        NSString *map;
        map = [letterMappingDict translateLetter: letter];

        [string appendString: map];
    }

    [self setOriginalString: string];

} // copyReplacementLettersToCryptogram



// The 'enter cryptogram' sheet has been dismissed.  If the user OK'd
// replace our current cryptogram with a new one

- (void) sheetDidEnd: (NSWindow *) sheet
          returnCode: (int) returnCode
         contextInfo: (void*) contextInfo
{
    if (returnCode == NSOKButton) {
        NSString *string;
        string = [stringTextField stringValue];

        [self setOriginalString: string];

        [self updateLetterFrequencies];
    }

    [stringEntrySheet orderOut: self];

} // sheetDidEnd



// menu item handler for "Enter Cryptogram".  This drops a sheet
// for the user to fill in.  It gets populated with the original
// string, so that the user can edit a cryptogram if they messed up
// the typing originally

- (IBAction) enterCryptogram: (id) sender
{
    [stringTextField setStringValue: originalString];

    [NSApp beginSheet: stringEntrySheet
           modalForWindow: [self window]
           modalDelegate: self
           didEndSelector: @selector(sheetDidEnd:returnCode:contextInfo:)
           contextInfo: NULL];
    
} // enterCryptogram



// OK button handler for the sheet

- (IBAction) entrySheetOK: (id) sender
{
    [NSApp endSheet: stringEntrySheet
           returnCode: NSOKButton];

} // entrySheetOK



// Cancel button handler for the sheet

- (IBAction) entrySheetCancel: (id) sender
{
    [NSApp endSheet: stringEntrySheet
           returnCode: NSCancelButton];

} // entrySheetCancel



// undo bottleneck for changing a letter translation

- (void) updateTranslatedLetter: (NSString *) letter
              forOriginalLetter: (NSString *) original
{
    NSString *oldLetter;

    oldLetter = [letterMappingDict translateLetter: original];

    NSUndoManager *undoManager = [self undoManager];

    [[undoManager prepareWithInvocationTarget: self]
        updateTranslatedLetter: oldLetter
        forOriginalLetter: original];

    [undoManager setActionName: @"Change Letter"];

    [letterMappingDict setTranslatedLetter: letter
                       forOriginalLetter: original];

    [cryptoView setNeedsDisplay: YES];
    
} // updateTranslatedLetter



// the view things that a letter mapping should be changed. 

- (void) handleLetterMapChangeNotification: (NSNotification *) notification
{
    NSDictionary *userInfo;

    userInfo = [notification userInfo];

    NSString *originalLetter;
    originalLetter = [userInfo objectForKey: BWCryptoView_OriginalLetterKey];

    NSString *newLetter;
    newLetter = [userInfo objectForKey: BWCryptoView_NewLetterKey];
    
    [self updateTranslatedLetter: newLetter
          forOriginalLetter: originalLetter];
    
    [cryptoView reloadData];
    [self updateLetterFrequencies];

} // handleLetterMapChangeNotification


@end // BWCryptoDocument



