// BWCryptoView.m -- implementaiton of the view that will draw the string
//		     as a bunch of blocks, with the replacement letters
//		     seated on top of the original letter.


#import "BWCryptoView.h"
#import "NSDictionary+Letter.h"


// this notification is posted when the user changes a letter

NSString *BWCryptoView_LetterMapChangeNotification
    = @"BWCryptoView Letter Map Change Notification";

// in the notification userInfo dictionary, the original letter and
// replacement letter (as NSStrings) are stored under these keys

NSString *BWCryptoView_OriginalLetterKey
    = @"BWCryptoView Original Letter Key";
NSString *BWCryptoView_NewLetterKey
    = @"BWCryptoView New Letter Key";


// space characters are special, being used to break words when wrapping
// the character boxes.  This string is just a single space, but makes it
// easy to do an "isEqualTo:" with

static NSString *g_spaceString;


// magic number for having an empty selection

#define NO_BLOCK_SELECTED (-1)


// the size of the rectangles the letters are drawn in

#define RECT_SIZE (26)


// how much to skootch the blocks away from the view border

#define LAYOUT_MARGIN (5)




@implementation BWCryptoView


// the NSView designated initializer.  

- (id) initWithFrame: (NSRect) frame
{
    if ((self = [super initWithFrame: frame])) {
        needToLayOutBlocks = YES; // force a recalc on first draw

        selectedBlockIndex = NO_BLOCK_SELECTED; // we have nothing selected

        NSFont *font;

        font = [NSFont fontWithName: @"MarkerFelt-Wide"  size: 16];
        // how did I find this?  by using
        // NSLog (@"%@", [[NSFontManager sharedFontManager] 
        //                   availableFontNamesWithTraits: 0]);

        replacementLetterAttributes 
            = [NSDictionary 
                  dictionaryWithObjectsAndKeys:
                      font, NSFontAttributeName,
                  [NSColor blackColor], NSForegroundColorAttributeName,
                  nil];
        [replacementLetterAttributes retain];
        
        font = [NSFont fontWithName: @"Helvetica"  size: 16];
        letterAttributes 
            = [NSDictionary 
                  dictionaryWithObjectsAndKeys:
                      font, NSFontAttributeName,
                  [NSColor blackColor], NSForegroundColorAttributeName,
                  nil];
        [letterAttributes retain];
    }

    // we do a lot of comparing against spaces, so keep a single-space
    // string handy

    if (g_spaceString == nil) {
        char *space = " ";
        g_spaceString = [NSString stringWithCString: space];
        [g_spaceString retain];
    }

    return (self);
    
} // initWithFrame



// clean up our mess

- (void) dealloc
{
    [originalString release];
    [letterMappingDict release];
    [replacementLetterAttributes release];
    [letterAttributes release];

    free (blocks);

    [super dealloc];

} // dealloc



// is is the given one-character string a letter?  Assumes the character
// will be upper case, and not have any diacritical marks

- (BOOL) characterIsLetter: (NSString *) character
{
    BOOL izzit = YES;;

    NSRange range;
    range = [character rangeOfCharacterFromSet: 
                           [NSCharacterSet uppercaseLetterCharacterSet]];

    if (range.location == NSNotFound) {
        izzit = NO;
    }

    return (izzit);
    
} // characterIsLetter



// set the original, encrypted string.  Walk the string and make
// blocks for them.

- (void) setOriginalString: (NSString *) string
{
    [originalString release];
    originalString = [[string uppercaseString] retain];

    blockCount = [originalString length];

    if (blocks != NULL) {
        free (blocks);
    }
    blocks = calloc (blockCount, sizeof(LetterBlock));
    
    int i;
    for (i = 0; i < blockCount; i++) {
        NSRange range = NSMakeRange (i, 1);
        blocks[i].letter = [[originalString substringWithRange: range] retain];
    }

    selectedBlockIndex = NO_BLOCK_SELECTED;

    [self setNeedsDisplay: YES];
    needToLayOutBlocks = YES;

} // setOriginalString



// return our original, encrypted string

- (NSString *) originalString
{
    return (originalString);
} // string



// this is our new letter mapping dictionary, which has original
// letters as the keys, and the replacement letters as the values

- (void) setLetterMappingDict: (NSDictionary *) dictionary
{
    [dictionary retain];
    [letterMappingDict release];
    letterMappingDict = dictionary;
    [self setNeedsDisplay: YES];

} // setLetterMappingDict



// return the letter mapping dictionary we're using

- (NSDictionary *) letterMappingDict
{
    return (letterMappingDict);
} // letterMappingDict



// someone has changed something (like the letter mapping dict).  Redisplay
// ourselves

- (void) reloadData
{
    [self setNeedsDisplay: YES];
} // reloadData



// word-wrap the string so that it fits horizontally in the rect.
// Long strings could go off the bottom of the rectangle.

- (void) layoutStringInBounds: (NSRect) rect
{
    // start filling in blocks from left to right

    // how much horizontal space we have to work with
    float width;
    width = rect.size.width - (2 * LAYOUT_MARGIN);

    // this is the point where to anchor the next rectangle
    NSPoint origin = NSMakePoint (rect.origin.x + LAYOUT_MARGIN,
                                  rect.origin.y + LAYOUT_MARGIN);

    // spaces are used to break words.  Remember where the last space
    // was.  if we run out of horizontal space, rewind to the last
    // space character found, and then move to a new line
    int lastSpaceIndex = -1;

    NSSize size = NSMakeSize(RECT_SIZE, RECT_SIZE);

    // this is the accumulated width of the line so far
    float widthSum = LAYOUT_MARGIN;

    // this is the index of the block at the beginning of the line.
    // if a word has to wrap, and we haven't seen a space so far
    // on this line, then we'll have to break the word in the middle
    int beginningOfLine = 0;

    // walk each block and find them a home
    int i;
    for (i = 0; i < blockCount; i++) {

        // the replacement letter rectangle sits on top of the original
        // letter rectangle
        blocks[i].replacementLetterRect.origin = origin;
        blocks[i].replacementLetterRect.size = size;

        // stick the letter rect immediately below the replacement letter
        blocks[i].letterRect.origin = NSMakePoint (origin.x, 
                                                   origin.y + RECT_SIZE);
        blocks[i].letterRect.size = size;

        // remember if we're a space, in case we need to break the line
        // here
        if ([blocks[i].letter isEqualTo: g_spaceString]) {
            lastSpaceIndex = i;
        }

        // calcuate the running total.
        widthSum += RECT_SIZE;

        if (widthSum > width) {
            // we wrapped

            if (lastSpaceIndex < beginningOfLine) {
                // the current line is too big to fit in the window,
                // break it in the middle
                // don't back up, just wrap right here
            } else {
                i = lastSpaceIndex;
                // at the end of the loop, i will be increment,
                // which neatly skips over the space and starts the
                // next character all the way to the left
            }

            // reset the width accumulation 
            widthSum = LAYOUT_MARGIN;

            // move to the beginning of the next line
            origin.x = rect.origin.x + LAYOUT_MARGIN;
            origin.y += (RECT_SIZE * 2) + RECT_SIZE / 2.0;

            // this is where the current start of line is
            beginningOfLine = i + 1;

        } else {
            // otherwise, just scoot over horizontally to position the
            // next rectangle
            origin.x += RECT_SIZE;
        }
    }

} // layoutStringInBounds



// draw a letter centered in a rectangle, and draw it using the given
// attributes.  draw a box around the letter if it's actually a letter
// the user can change

- (void) drawLetter: (NSString*) letter  inRect: (NSRect) rect
     withAttributes: (NSDictionary *) attributes
{
    // draw a box if it's a letter
    if ([self characterIsLetter: letter]) {
        // scootch the origin over a bit so that we get single-pixel lines
        rect.origin.x += 0.5;
        rect.origin.y += 0.5;
        [NSBezierPath strokeRect: rect];
    }
    
    // get size of letter
    NSSize letterSize = [letter sizeWithAttributes: attributes];
    
    // figure out the point where to place the letter.  The bottom-left
    // of the letter will be going at this location, so scoot
    // back from the middle of the rectangle by half the width and height
    NSPoint point;
    point.x = rect.origin.x + (rect.size.width / 2 - (letterSize.width / 2));
    point.y = rect.origin.y + (rect.size.height / 2 - (letterSize.height / 2));
    
    // draw that pupper
    [letter drawAtPoint: point  withAttributes: attributes];
    
} // drawLetter



// draw the whole cryptoview thingie

- (void) drawRect: (NSRect) rect
{
    // go ahead and just draw everything.
    rect = [self bounds];

    if (needToLayOutBlocks) {
        [self layoutStringInBounds: rect];
        needToLayOutBlocks = NO;
    }

    // clear out to a nice white color
    [[NSColor whiteColor] set];
    [NSBezierPath fillRect: rect];

    // find out what the selected letter is.  Draw all the letters that
    // match the selected letter in a light gray, and draw the actual
    // selected letter in dark gray
    NSString *selectedLetter = nil;

    if (selectedBlockIndex != NO_BLOCK_SELECTED) {
        selectedLetter = blocks[selectedBlockIndex].letter;
    }

    // spin through each of the blocks and draw them.
    // this is actually pretty easy since we've already calculated
    // their bounding rectangles
    int i;
    for (i = 0; i < blockCount; i++) {

        // don't draw anything for spaces
        if (![blocks[i].letter isEqualTo: g_spaceString]) {
            
            // make the selected block darker
            if (selectedBlockIndex == i) {
                [[NSColor grayColor] set];
                [NSBezierPath fillRect: blocks[i].replacementLetterRect];

            } else if ([blocks[i].letter isEqualTo: selectedLetter]) {
                // and make letters that match the selected block a little
                // darker, so they stand out
                [[NSColor colorWithCalibratedWhite: 0.85  alpha: 1.0] set];
                [NSBezierPath fillRect: blocks[i].replacementLetterRect];
            }

            // now draw everything else in black
            [[NSColor blackColor] set];

            // figure out what the letter maps to
            NSString *translatedLetter;
            translatedLetter = [letterMappingDict 
                                   translateLetter: blocks[i].letter];

            [self drawLetter: translatedLetter
                  inRect: blocks[i].replacementLetterRect
                  withAttributes: replacementLetterAttributes];

            [self drawLetter: blocks[i].letter
                  inRect: blocks[i].letterRect
                  withAttributes: letterAttributes];
        }
    }
    
    // put a nice frame around everything
    [[NSColor blackColor] set];
    [NSBezierPath strokeRect: rect];

} // drawRect



// easy way to find when we get resized.  If we do get resized, we need
// to recalculate the block positions

- (void) setFrame: (NSRect) frameRect
{
    [super setFrame: frameRect];
    needToLayOutBlocks = YES;

} // setFrame



// we're aflippa da bits.  I think better with an upper-left origin

- (BOOL) isFlipped
{
    return (YES);
} // isFlipped


- (void) setNeedsDisplayForLetter: (NSString *) letter
{
    int i;
    for (i = 0; i < blockCount; i++) {
        if ([letter isEqualTo: blocks[i].letter]) {
            [self setNeedsDisplayInRect: blocks[i].replacementLetterRect];
        }
    }

} // setNeedsDisplayForLetter


// given a point, see if it hits a replacement letter box.  if
// if does, make that box the selected block

- (void) hitTestPoint: (NSPoint) point
{
    // see which block it is
    int newSelectedBlock = NO_BLOCK_SELECTED;

    int i;
    for (i = 0; i < blockCount; i++) {

        if ([self characterIsLetter: blocks[i].letter]) {
            if (NSPointInRect(point, blocks[i].replacementLetterRect)) {
                newSelectedBlock = i;
                break;
            }
        }
    }

    // only redraw on a change
    if (newSelectedBlock != selectedBlockIndex) {
        // only redraw the letters in question
        if (newSelectedBlock != NO_BLOCK_SELECTED) {
            [self setNeedsDisplayForLetter: blocks[newSelectedBlock].letter];
        }
        if (selectedBlockIndex != NO_BLOCK_SELECTED) {
            [self setNeedsDisplayForLetter: blocks[selectedBlockIndex].letter];
        }
        selectedBlockIndex = newSelectedBlock;
    }

} // hitTestPoint



// the user clicked.  If they click in a replacement letter, highlight it.

- (void) mouseDown: (NSEvent *) event
{
    NSPoint mouse;
    mouse = [self convertPoint: [event locationInWindow]  fromView: nil];

    [self hitTestPoint: mouse];

} // mouseDown



// we're in a mouse drag.  if they're over a replacement letter, highlight
// it.

- (void) mouseDragged: (NSEvent *) event
{
    NSPoint mouse;
    mouse = [self convertPoint: [event locationInWindow]  fromView: nil];

    [self hitTestPoint: mouse];

} // mouseDragged



// move the selected block left or right.  Wrap around (like if they moved
// left when at position zero).  Skip over blocks that aren't letters.

- (void) moveSelectedBlockBy: (int) delta
{
    // can only really move if there are two or more blocks
    if (blockCount > 1) {

        while (1) {

            // move by the given delta

            selectedBlockIndex += delta;
            
            // handle wrap-around
            if (selectedBlockIndex < 0) {
                selectedBlockIndex = blockCount - 1; // wrap around
            }
            if (selectedBlockIndex >= blockCount) {
                selectedBlockIndex = 0;
            }

            // if we're on a letter, we can stop, otherwise keep spinning
            // until we find a ltter

            if ([self characterIsLetter: blocks[selectedBlockIndex].letter]) {
                // we're on a letter we can change
                [self setNeedsDisplay: YES];
                break;
            }
        }
    }

} // moveSelectedBlockBy



// handle a key press.

- (void) keyDown: (NSEvent *) event
{
    NSString *characters;
    characters = [event characters];

    // only do work if there's a selected block

    if (selectedBlockIndex != NO_BLOCK_SELECTED) {
        
        unichar character;
        character = [characters characterAtIndex: 0];

        // see if we should move the selected block left or right
        if (character == NSRightArrowFunctionKey) {
            [self moveSelectedBlockBy: 1];

        } else if (character == NSLeftArrowFunctionKey) {
            [self moveSelectedBlockBy: -1];
            
        } else {

            // otherwise handle a letter key press.

            NSString *newLetter;
            newLetter = [NSString stringWithCharacters: &character
                                  length: 1];

            // bash it to upper case
            newLetter = [newLetter uppercaseString];

            // make sure it's actually a letter before doing anything
            if ([self characterIsLetter: newLetter]) {

                // ok, we passed all the hurdles.  Post a notificaiton
                // about the change.  Someone else will tweak the 
                // the letterMappignDictionary
                NSDictionary *userInfo;
                userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
                                             newLetter,
                                         BWCryptoView_NewLetterKey,
                                         blocks[selectedBlockIndex].letter,
                                         BWCryptoView_OriginalLetterKey,
                                         nil];

                NSNotificationCenter *center;
                center = [NSNotificationCenter defaultCenter];

                [center postNotificationName: 
                            BWCryptoView_LetterMapChangeNotification
                        object: self
                        userInfo: userInfo];
            }
        }
    }

} // keyDown



// we accept first responder so we can accept keypresses

- (BOOL) acceptsFirstResponder
{
    return (YES);
} // return acceptsFirstResponder



@end // BWCryptoView

