// NSDictionary+Letter.h -- category on NS[Mutable]Dictionary to handle the 
//                          letter translation stuff.

// this turned out to be a very elegant way to address this problem.
// Both the document and the cryptoview had to query the dictionary.
// I could either put the letter mapping dictionary on the view, but that
// made saving the document a little strange.  I could put it on the
// document, but then the view would have to know about the document in
// order to map letters when drawing.  By moving the code to a category
// on NS[Mutable]Dictionary, both classes can now access this code, without
// any design contortions or duplicated code.


#import "NSDictionary+Letter.h"

@implementation NSDictionary(LetterTranslation)

// given a letter, return the mapping.  If there is no mapping, just
// return the passed in letter.

- (NSString *) translateLetter: (NSString *) letterString
{
    NSString *replacement;
    replacement = [self objectForKey: letterString];

    // if not found, just return as itself
    if (replacement == nil) {
        replacement = letterString;
    }

    return (replacement);

} // translateLetter

@end


@implementation NSMutableDictionary(LetterTranslation)


// throw out all the mappings, and replace it with "A=A, B=B, C=C ..."

- (void) resetTranslationToIdentity
{
    // start from an empty slate
    [self removeAllObjects];

    // populate with identity
    char scan;
    for (scan = 'A'; scan <= 'Z'; scan++) {
        NSString *letter;
        letter = [NSString stringWithCString: &scan  length: 1];
        [self setObject: letter  forKey: letter];
    }

} // resetTranslationToIdentity



// change the mapping so that 'original' mapps to 'letter'

- (void) setTranslatedLetter: (NSString *) letter
           forOriginalLetter: (NSString *) original
{
    // the actual work is a little more complicated.  Say the user is
    // wanting "N" (original) to have "A" (letter)
    // "N" Currently maps to "X" (currentLetter)
    // we want the guy who is currently mapped to "A" (which is "R") 
    //     to map to "X".
    // So, set "R" to map to "X"
    // then set "N" to map to "A"

    NSString *currentLetter;
    currentLetter = [self translateLetter: original];

    NSArray *keys;
    keys = [self allKeysForObject: letter];

    NSString *inverse;
    inverse = [keys objectAtIndex: 0];

    [self setObject: currentLetter  forKey: inverse];
    [self setObject: letter  forKey: original];

} // setTranslatedLetter


@end

